/*************************************************************************

Copyright (c) 1994  Microsoft Corporation

Module Name:

	dibfilt.cpp

Abstract:

	Some signal processing APIs for dib:
	2-D median filter: Median2D. NOTE: odd window size only.
	1-DX median filter: Median1DX. NOTE: odd window size only.
	1-DY median filter: Median1DX. NOTE: odd window size only.
	1-DX linear filter: Transversal1DX. NOTE: odd filter length only.
	1-DY linear filter: Transversal1DY. NOTE: odd filter length only.
	2-D linear separable filter: Transversal2D. NOTE: odd filter length only.

References: 

	Handbook of Digital Signal Processing, Mitra & Kaiser.  Pg 1029 for 
		filter coeffs.
	Multirate Systems and Filter Banks, Vaidyananathan, chapter 4.
	Two Dimensional Signal and Image Processing by Lim.

Author:

	Ming-Chieh Lee (MingCL) 11-November-1994

Revision History:

*************************************************************************/

#include "stdafx.h"
#include "dib.h"
#include "dibfilt.h"

inline void swap (BYTE& a, BYTE& b)
{
	BYTE tmp = a;
	a = b;
	b = tmp;
}


inline BYTE select (int k, int size, BYTE* data)
{
	int left = 0, right = size - 1;
	for (;;) 
	{
		if (right <= left + 1) 
		{
			if (right == left + 1 && data [right] < data [left])
			{ 
				swap (data [left], data [right]);
			}
			return data [k];
		} 
		else 
		{
			int mid = (left + right) >> 1;
			swap (data [mid], data [left + 1]);
			if (data [left + 1] > data [right]) 
			{
				swap (data [left + 1], data [right]);
			}
			if (data [left] > data [right]) 
			{
				swap (data [left], data [right]);
			}
			if (data [left + 1] > data [left]) 
			{
				swap (data [left + 1], data [left]);
			}
			int i = left + 1;
			int j = right;
			BYTE a = data [left];
			for (;;) 
			{
				do i++; while (data [i] < a);
				do j--; while (data [j] > a);
				if (j < i) break;
				swap (data [i], data [j]);
			}
			data [left] = data [j];
			data [j] = a;
			if (j >= k) 
			{
				right = j - 1;
			}
			if (j <= k) 
			{
				left = i;
			}
		}
	}
}


PDIB Median2D (const PDIB pDib, int window)
{
	PDIB pRetDib = DibCopy (pDib);	// return dib
	const BYTE* ppdib = (const BYTE*) DibPtr (pDib);
	BYTE* ppret = (BYTE*) DibPtr (pRetDib);

	const int offset = window >> 1; // boundary not to be filtered
	const int offset2 = (offset << 1) * 3;
	const int size = window * window; // array size to be sorted
	const int med = size >> 1;
	const int width = (int) DibWidth (pDib);
	const int height = (int) DibHeight (pDib);
	// boundary of the image to be filtered.
	const int left = offset;
	const int top = offset;
	const int right = width - offset;
	const int bottom = height - offset;
	const int dist = (offset + width) * 3;
	const int wwidth = (width - window) * 3;
	register BYTE* r = new BYTE [size];	// temporary memory for sorting
	register BYTE* g = new BYTE [size];
	register BYTE* b = new BYTE [size];
	BYTE* pret = ppret + 3 * (left + width * top);
	const BYTE* pv = ppdib + 3 * (left + width * top);
	for (int y = top; y != bottom; y++)
	{
		for (int x = left; x != right; x++)
		{
			const BYTE* pp = pv - dist; // get index of the starting position of the window
			register BYTE* rr = r;
			register BYTE* gg = g;
			register BYTE* bb = b;
			for (int sy = 0; sy != window; sy++)
			{
				for (int sx = 0; sx != window; sx++)
				{
					*bb++ = *pp++; // bgr
					*gg++ = *pp++;
					*rr++ = *pp++;
				}
				pp += wwidth;
			}
			*pret++ = select (med, size, b); // select the median one
			*pret++ = select (med, size, g); 
			*pret++ = select (med, size, r); 
			pv += 3;
		}
		pret += offset2;
		pv += offset2;
	}
	delete [] r;
	delete [] g;
	delete [] b;
	return pRetDib;
}


PDIB Median1DX (const PDIB pDib, int window)
{
	PDIB pRetDib = DibCopy (pDib);	// return dib
	const BYTE* ppdib = (const BYTE*) DibPtr (pDib);
	BYTE* ppret = (BYTE*) DibPtr (pRetDib);

	const int offset = window >> 1;
	const int offset2 = (offset << 1) * 3;
	const int dist = offset * 3;
	const int size = window; // array size to be sorted
	const int med = size >> 1;
	const int width = (int) DibWidth (pDib);
	const int height = (int) DibHeight (pDib);
	// bound of the image to be filtered.
	const int left = offset;
	const int top = 0;
	const int right = width - offset;
	const int bottom = height;
	register BYTE* r = new BYTE [size];	// faster if use register??
	register BYTE* g = new BYTE [size];
	register BYTE* b = new BYTE [size];
	BYTE* pret = ppret + 3 * (left + width * top);
	const BYTE* pv = ppdib + 3 * (left + width * top);
	for (int y = top; y != bottom; y++)
	{
		for (int x = left; x != right; x++)
		{
			register BYTE* rr = r;
			register BYTE* gg = g;
			register BYTE* bb = b;
			const BYTE* pp = pv - dist;
			for (int s = 0; s != size; s++)
			{
				*bb++ = *pp++;
				*gg++ = *pp++;
				*rr++ = *pp++;
			}
			*pret++ = select (med, size, b); // select the median one
			*pret++ = select (med, size, g); 
			*pret++ = select (med, size, r); 
			pv += 3;
		}
		pret += offset2;
		pv += offset2;
	}
	delete [] r;
	delete [] g;
	delete [] b;
	return pRetDib;
}


PDIB Median1DY (const PDIB pDib, int window)
{
	PDIB pRetDib = DibCopy (pDib);	// return dib
	const BYTE* ppdib = (const BYTE*) DibPtr (pDib);
	BYTE* ppret = (BYTE*) DibPtr (pRetDib);

	const int offset = window >> 1;
	const int size = window; // array size to be sorted
	const int med = size >> 1;
	const int width = (int) DibWidth (pDib);
	const int width3 = width * 3 - 3;
	const int dist = width * offset * 3;
	const int height = (int) DibHeight (pDib);
	// bound of the image to be filtered.
	const int left = 0;
	const int top = offset;
	const int right = width;
	const int bottom = height - offset;
	register BYTE* r = new BYTE [size];
	register BYTE* g = new BYTE [size];
	register BYTE* b = new BYTE [size];
	BYTE* pret = ppret + 3 * (left + width * top);
	const BYTE* pv = ppdib + 3 * (left + width * top);
	for (int y = top; y != bottom; y++)
	{
		for (int x = left; x != right; x++)
		{
			register BYTE* rr = r;
			register BYTE* gg = g;
			register BYTE* bb = b;
			const BYTE* pp = pv - dist;
			for (int s = 0; s != size; s++)
			{
				*bb++ = *pp++;
				*gg++ = *pp++;
				*rr++ = *pp++;
				pp += width3;
			}
			*pret++ = select (med, size, b); // select the median one
			*pret++ = select (med, size, g); 
			*pret++ = select (med, size, r); 
			pv += 3;
		}
	}
	delete [] r;
	delete [] g;
	delete [] b;
	return pRetDib;
}


PDIB Transversal1DX	(const PDIB pDib, int length, const double* coeff)
{
	PDIB pRetDib = DibCopy (pDib);	// return dib
	const BYTE* ppdib = (const BYTE*) DibPtr (pDib);
	BYTE* ppret = (BYTE*) DibPtr (pRetDib);

	const int offset = length >> 1;
	const int offset2 = (offset << 1) * 3;
	const int width = (int) DibWidth (pDib);
	const int height = (int) DibHeight (pDib);
	// bound of the image to be filtered.
	const int left = offset;
	const int top = 0;
	const int right = width - offset;
	const int bottom = height;
	const double* cfs = coeff + offset;
	BYTE* pret = ppret + 3 * (left + width * top);
	const BYTE* pv = ppdib + 3 * (left + width * top);
	for (int y = top; y != bottom; y++)
	{
		for (int x = left; x != right; x++)
		{
			register double r = 0, g = 0, b = 0;
			const BYTE* plat = pv;
			const BYTE* ppre = pv;
			// Take advantage of the symmetry to do convolution
			for (int k = 1; k <= offset; k++)
			{
				register double ceff = cfs [k];
				ppre -= 3;
				plat += 3;
				b += ceff * (*plat + *ppre);
				g += ceff * (*(plat + 1) + *(ppre + 1));
				r += ceff * (*(plat + 2) + *(ppre + 2));
			}
			b += *cfs * *pv++;
			b = max (b, 0);
			*pret++ = (BYTE) min (b, 255);
			g += *cfs * *pv++;
			g = max (g, 0);
			*pret++ = (BYTE) min (g, 255);
			r += *cfs * *pv++;
			r = max (r, 0);
			*pret++ = (BYTE) min (r, 255);
		}
		pv += offset2;
		pret += offset2;
	}
	return pRetDib;
}


PDIB Transversal1DY	(const PDIB pDib, int length, const double* coeff)
{
	PDIB pRetDib = DibCopy (pDib);	// return dib
	const BYTE* ppdib = (const BYTE*) DibPtr (pDib);
	BYTE* ppret = (BYTE*) DibPtr (pRetDib);

	const int offset = length >> 1;
	const int width = (int) DibWidth (pDib);
	const int dist = width * 3;
	const int height = (int) DibHeight (pDib);
	// bound of the image to be filtered.
	const int left = 0;
	const int top = offset;
	const int right = width;
	const int bottom = height - offset;
	const double* cfs = coeff + offset;
	BYTE* pret = ppret + 3 * (left + width * top);
	const BYTE* pv = ppdib + 3 * (left + width * top);
	for (int y = top; y != bottom; y++)
	{
		for (int x = left; x != right; x++)
		{
			register double r = 0, g = 0, b = 0;
			const BYTE* plat = pv;
			const BYTE* ppre = pv;
			// Take advantage of the symmetry to do convolution
			for (int k = 1; k <= offset; k++)
			{
				register double ceff = cfs [k];
				ppre -= dist;
				plat += dist;
				b += ceff * (*plat + *ppre);
				g += ceff * (*(plat + 1) + *(ppre + 1));
				r += ceff * (*(plat + 2) + *(ppre + 2));
			}
			b += *cfs * *pv++;
			b = max (b, 0);
			*pret++ = (BYTE) min (b, 255);
			g += *cfs * *pv++;
			g = max (g, 0);
			*pret++ = (BYTE) min (g, 255);
			r += *cfs * *pv++;
			r = max (r, 0);
			*pret++ = (BYTE) min (r, 255);
		}
	}
	return pRetDib;
}

PDIB Transversal2D (const PDIB pDib, int lengthX, const double* coeffX, int lengthY, const double* coeffY)
{
	PDIB pFiltX = Transversal1DX (pDib, lengthX, coeffX);
	PDIB pFiltXY = Transversal1DY (pFiltX, lengthY, coeffY);
	DibFree (pFiltX);
	return pFiltXY;
}
